# SPDX-FileCopyrightText: 2023 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# bindings.py and accompanying platform files are generated and emplaced automatically using pypdfium2 setup tooling - see autorelease/bindings.py for a tracked sample
from pypdfium2_raw.bindings import *



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2._cli._parsers import add_input, get_input

EXTRACT_RANGE   = "range"
EXTRACT_BOUNDED = "bounded"

# __main__.py hook
PARSER_DESC = """\
Note that PDFium outputs CRLF (\\r\\n) style line breaks.
This may be undesirable or confusing in some situations, e.g. when processing the output with an (unaware) parser on the command line.
If this is an issue, run e.g. `dos2unix` on the output, or use the Python API.\
"""

def attach(parser):
    add_input(parser, pages=True)
    parser.add_argument(
        "--strategy",
        default = EXTRACT_RANGE,
        choices = (EXTRACT_RANGE, EXTRACT_BOUNDED),
        help = "PDFium text extraction strategy (range, bounded).",
    )


def main(args):
    
    pdf = get_input(args)
    
    sep = ""
    for i in args.pages:
        
        page = pdf[i]
        textpage = page.get_textpage()
        
        # TODO let caller pass in possible range/boundary parameters
        if args.strategy == EXTRACT_RANGE:
            text = textpage.get_text_range()
        elif args.strategy == EXTRACT_BOUNDED:
            text = textpage.get_text_bounded()
        else:
            assert False
        
        print(sep + f"# Page {i+1}\n" + text)
        sep = "\n"



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import shlex
import argparse
import traceback

# local
from base import *
import update as update_pdfium
import build_native
import build_toolchained
import system_pdfium


def _repr_info(version, flags):
    return str(version) + (f":{','.join(flags)}" if flags else "")

def _get_pdfium_with_cache(pl_name, req_ver, req_flags):
    
    # TODO turn platform and system into proper objects, so the libname could be accessed like plat.system.libname, which is much cleaner than a chain of string function calls
    
    pl_dir = DataDir/pl_name
    system = plat_to_system(pl_name)
    binary = pl_dir/libname_for_system(system)
    binary_ver = pl_dir/VersionFN
    
    if all(f.exists() for f in (binary, binary_ver)):
        prev_info = read_json(binary_ver)
        update_binary = prev_info["build"] != req_ver or set(prev_info["flags"]) != set(req_flags)
    else:
        update_binary = True
    
    req_repr = _repr_info(req_ver, req_flags)
    if update_binary:
        log(f"Downloading binary {req_repr} ...")
        update_pdfium.main([pl_name], version=req_ver, use_v8=("V8" in req_flags))
    else:
        log(f"Using cached binary {req_repr}")
    
    # build_pdfium_bindings() has its own cache logic, so always call to ensure bindings match
    ct_paths = (DataDir/Host.platform/CTG_LIBPATTERN, ) if pl_name == Host.platform else ()
    build_pdfium_bindings(req_ver, flags=req_flags, ct_paths=ct_paths)

def _end_subtargets(sub_target, pdfium_ver):
    if sub_target:
        assert False, sub_target
    else:
        log("No sub-target set, will use existing data files.")
        if pdfium_ver:
            raise ValueError(f"Pdfium version {pdfium_ver} was passed, but this does not make sense with caller-provided data files.")


def stage_platfiles(pl_name, sub_target, pdfium_ver, flags):
        
    if pl_name == ExtPlats.system:
        pl_dir = DataDir/pl_name
        if sub_target:
            purge_dir(pl_dir)
        if sub_target == "search":
            full_ver = PdfiumVer.to_full(pdfium_ver) if pdfium_ver else None
            full_ver = system_pdfium.main(full_ver, flags=flags)
        elif sub_target == "generate":
            assert pdfium_ver, "system-generate target requires pdfium build version from caller"
            build_pdfium_bindings(pdfium_ver, flags=flags, guard_symbols=True, rt_paths=())
            shutil.copyfile(BindingsFile, pl_dir/BindingsFN)
            full_ver = PdfiumVer.to_full(pdfium_ver)
            write_pdfium_info(pl_dir, full_ver, origin="system-generate", flags=flags)
        else:
            _end_subtargets(sub_target, pdfium_ver)
    
    elif pl_name == ExtPlats.sourcebuild:
        if flags:
            log(f"sourcebuild: flags {flags!r} are not handled (will be discarded).")
        
        build_params = shlex.split( os.getenv("BUILD_PARAMS", "") )
        builder = dict(native=build_native, toolchained=build_toolchained).get(sub_target)
        if builder:
            build_params = vars(builder.parse_args(build_params))
            build_params.update(dict(build_ver=pdfium_ver))
            log(build_params)
            builder.main(**build_params)
        else:
            _end_subtargets(sub_target, pdfium_ver)
    
    elif pl_name == ExtPlats.fallback:
        pl_name = ExtPlats.system
        try:
            stage_platfiles(pl_name, "search", pdfium_ver, flags)
        except system_pdfium.PdfiumNotFoundError:
            log("Could not find system pdfium, will attempt native sourcebuild")
            pl_name = ExtPlats.sourcebuild
            try:
                stage_platfiles(pl_name, "native", pdfium_ver, flags)
            except Exception:
                traceback.print_exc()
                raise RuntimeError("sourcebuild-native failed. Manual action may be needed, such as installing system dependencies, or possibly patching the sources. See pypdfium2's README.md for more information.")
    
    else:
        if not pdfium_ver or pdfium_ver == "pinned":
            pdfium_ver = PdfiumVer.pinned
            log(f"Using pinned pdfium version {pdfium_ver!r}. If this is not intentional, set e.g. {PlatSpec_EnvVar}=auto:latest to use the latest version instead.")
        elif pdfium_ver == "latest":
            pdfium_ver = PdfiumVer.get_latest()
            log(f"Using latest pdfium-binaries version {pdfium_ver!r}.")
        assert pl_name and hasattr(PlatNames, pl_name)
        _get_pdfium_with_cache(pl_name, pdfium_ver, flags)
    
    return pl_name


def copy_platfiles(pl_name):
    
    # remove existing in-tree platform files, if any
    clean_platfiles()
    
    # the version file is in the platform directory for all targets
    pl_dir = DataDir/pl_name
    platfiles = [pl_dir/VersionFN]
    
    # For system and sourcebuild, the bindings file is in the platform directory.
    # For the pdfium-binaries targets, the bindings are shared in data/bindings/
    if pl_name == ExtPlats.system:
        platfiles.append(pl_dir/BindingsFN)
    elif pl_name == ExtPlats.sourcebuild:
        platfiles.append(pl_dir/BindingsFN)
        platfiles.extend(p for p in pl_dir.glob(Host.libname_glob))
    else:
        platfiles.append(BindingsFile)
        system = plat_to_system(pl_name)
        platfiles.append(pl_dir/libname_for_system(system))
    
    assert all(fp.exists() for fp in platfiles), "Some platform files are missing"
    for fp in platfiles:
        shutil.copy(fp, ModuleDir_Raw/fp.name)
    
    return tuple(fp.name for fp in platfiles)


def prepare_setup(pl_name, sub_target, pdfium_ver, flags):
    # Write platform files into a data staging directory
    pl_name = stage_platfiles(pl_name, sub_target, pdfium_ver, flags)
    # Copy platform files into actual source tree
    platfiles = copy_platfiles(pl_name)
    return platfiles, pl_name


def main():
    
    parser = argparse.ArgumentParser(
        description = "Manage in-tree artifacts from an editable install.",
    )
    parser.add_argument(
        "plat_spec",
        default = os.environ.get(PlatSpec_EnvVar, ""),
        nargs = "?",
        help = f"The platform specifier. Same format as of ${PlatSpec_EnvVar} on setup. {ExtPlats.sdist!r} removes existing artifacts.",
    )
    args = parser.parse_args()
    
    if args.plat_spec == ExtPlats.sdist:
        log("Remove existing in-tree platform files, if any.")
        clean_platfiles()
        return
    
    pl_name, *pl_info = parse_pl_spec(args.plat_spec)
    prepare_setup(pl_name, *pl_info)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# NOTE: This code is provided on a best effort basis. It is relatively hard for us to test.

import re
import sys
import shutil
import itertools
from pathlib import Path
from ctypes.util import find_library
from urllib.request import urlopen

from base import *  # local


def _get_existing(candidates, cb=Path.exists):
    return next((p for p in candidates if cb(p)), None)

def _removeprefix(string, prefix):
    if string.startswith(prefix):
        string = string[len(prefix):]
    return string


def _find_pdfium_headers():
    
    headers_path = os.getenv("PDFIUM_HEADERS")
    if headers_path:
        return Path(headers_path)
    
    if not sys.platform.startswith(("win", "darwin")):
        include_dirs = (Host.usr/"include", Host.usr/"local"/"include")
        candidates = (path/"pdfium" for path in include_dirs)
        headers_path = _get_existing(candidates, cb=Path.is_dir)
        if headers_path:
            return headers_path
        
        candidates = (path/"fpdf_edit.h" for path in include_dirs)
        sample_header = _get_existing(candidates, cb=Path.is_file)
        if sample_header:
            return sample_header.parent
    
    return None

def _parse_version(version):
    if "." in version:
        version = PdfiumVer.scheme(*[int(v) for v in version.split(".")])
    else:
        version = PdfiumVer.to_full(int(version))
    assert version.build > 1000, "unexpected versioning scheme, or grossly outdated"
    return version

def _get_sys_pdfium_ver(pdfium_lib):
    
    log("Trying to determine system pdfium version ...")
    
    libname_parts = Path(pdfium_lib).name.split(".", maxsplit=2)
    if len(libname_parts) == 3:
        return _parse_version(libname_parts[2])
    
    if shutil.which("pkg-config"):
        proc = run_cmd(["pkg-config", "--modversion", "libpdfium"], cwd=None, check=False)
        if proc.returncode == 0:
            version = proc.stdout.decode().strip()
            return _parse_version(version)
    
    log("Unable to identify version, will set NaN placeholders.")
    return PdfiumVerUnknown


def _yield_lo_candidates(system):
    lo_paths_iter = itertools.product(
        (Host.usr/"lib", Host.usr/"local"/"lib"), ("", "64")
    )
    libname = libname_for_system(system, name="pdfiumlo")
    yield from (
        Path(str(path)+bitness)/"libreoffice"/"program"/libname
        for path, bitness in lo_paths_iter
    )

def _find_lo_pdfium():
    # Look for pdfium bundled with libreoffice, assuming a unix-like filesystem hierarchy.
    pdfium_lib = None
    if not sys.platform.startswith(("win", "darwin")):
        candidates = _yield_lo_candidates(Host.system)
        pdfium_lib = _get_existing(candidates)
    return pdfium_lib

def _get_lo_pdfium_ver():
    log("Trying to determine libreoffice pdfium version ...")
    
    output = run_cmd(["libreoffice", "--version"], cwd=None, capture=True)
    # alternatively, we could do e.g.: re.search(r"([\d\.]+)", output)
    output = _removeprefix(output.lower(), "libreoffice").lstrip()
    lo_version = output.split(" ")[0]
    log(f"Libreoffice version: {lo_version!r}")
    
    deps_url = f"https://raw.githubusercontent.com/LibreOffice/core/refs/tags/libreoffice-{lo_version}/download.lst"
    deps_content = urlopen(deps_url).read().decode("utf-8")
    match = re.search(r"pdfium-(\d+)\.tar\.bz2", deps_content, flags=re.MULTILINE)
    short_ver = int(match.group(1))
    log(f"Libreoffice pdfium version: {short_ver}")
    
    return PdfiumVer.to_full(short_ver)


class PdfiumNotFoundError (RuntimeError):
    pass


def _yield_pdfium_candidates():
    # give the caller an opportunity to set the pdfium path
    yield os.getenv("PDFIUM_BINARY"), "caller"
    # see if a pdfium shared library is in the default system search path
    yield find_library("pdfium"), "ctypes"
    # see if libreoffice provides pdfium
    yield _find_lo_pdfium(), "libreoffice"

def _get_pdfium():
    candidates = _yield_pdfium_candidates()
    for pdfium_lib, finder in candidates:
        if pdfium_lib:
            return pdfium_lib, finder
    # abort if none of this worked
    raise PdfiumNotFoundError("Could not find system pdfium.")


def main(given_fullver=None, flags=(), target_dir=DataDir/ExtPlats.system):
    
    log("Looking for system pdfium ...")
    pdfium_lib, finder = _get_pdfium()
    
    log(f"Found pdfium shared library at {pdfium_lib} ({finder})")
    target_path = target_dir/BindingsFN
    kwargs = dict(univ_paths=(pdfium_lib,), guard_symbols=True, flags=flags)
    
    if finder == "libreoffice":
        full_ver = given_fullver or _get_lo_pdfium_ver()
        # assuming libreoffice does not change the original pdfium ABI
        build_pdfium_bindings(full_ver.build, **kwargs)
        bindings_path = BindingsFile
    else:
        pdfium_headers = _find_pdfium_headers()
        full_ver = given_fullver or _get_sys_pdfium_ver(pdfium_lib)
        if pdfium_headers:
            log(f"Found pdfium headers at {pdfium_headers}")
            run_ctypesgen(target_path, pdfium_headers, version=full_ver.build, **kwargs)
            bindings_path = target_path
        elif full_ver is not PdfiumVerUnknown:
            log(f"Could not find headers, but know the version: {full_ver}")
            build_pdfium_bindings(full_ver.build, **kwargs)
            bindings_path = BindingsFile
        else:
            log(f"Warning: Neither pdfium headers nor version found - will use reference bindings. This is ABI-unsafe! Set $PDFIUM_HEADERS to the directory in question, or pass the version via $PDFIUM_PLATFORM=system-search:$VERSION.")
            bindings_path = RefBindingsFile
    
    write_pdfium_info(target_dir, full_ver, origin=f"system-{finder}", flags=flags)
    if bindings_path != target_path:
        shutil.copyfile(bindings_path, target_path)
    if full_ver.build < PDFIUM_MIN_REQ:
        log(f"Warning: pdfium version {full_ver.build} does not conform with minimum requirement {PDFIUM_MIN_REQ}. Some APIs may not work. Run pypdfium2's test suite for details.")
    
    return full_ver


if __name__ == "__main__":
    # print(_get_lo_pdfium_ver())
    # print(_find_pdfium_headers())
    # print(_get_sys_pdfium_ver("libpdfium.so.140.0.7295.0"))
    print(main())



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause
# Related work: https://github.com/tiran/libpdfium and https://aur.archlinux.org/packages/libpdfium-nojs

import re
import os
import sys
import shutil
import argparse
from enum import Enum
from pathlib import Path

from base import *  # local

IS_CIBUILDWHEEL = bool(int( os.environ.get("CIBUILDWHEEL", 0) ))

PDFIUM_URL = "https://pdfium.googlesource.com/pdfium"
_CR_PREFIX = "https://chromium.googlesource.com/"
DEPS_URLS = dict(
    build      = _CR_PREFIX + "chromium/src/build",
    abseil     = _CR_PREFIX + "chromium/src/third_party/abseil-cpp",
    fast_float = _CR_PREFIX + "external/github.com/fastfloat/fast_float",
    catapult   = _CR_PREFIX + "catapult",  # android
    icu        = _CR_PREFIX + "chromium/deps/icu",  # for cibuildwheel
    gtest      = _CR_PREFIX + "external/github.com/google/googletest",
    test_fonts = _CR_PREFIX + "chromium/src/third_party/test_fonts",
)
SOURCES_DIR = ProjectDir / "sbuild" / "native"
PDFIUM_DIR = SOURCES_DIR / "pdfium"
PDFIUM_3RDPARTY = PDFIUM_DIR / "third_party"

Compiler = Enum("Compiler", "gcc clang")

DefaultConfig = {
    "is_debug": False,
    "use_glib": False,
    "use_remoteexec": False,
    "treat_warnings_as_errors": False,
    "clang_use_chrome_plugins": False,
    "is_component_build": False,
    "pdf_is_standalone": True,
    "pdf_enable_v8": False,
    "pdf_enable_xfa": False,
    "pdf_use_skia": False,
    "pdf_use_partition_alloc": False,
    # "sysroot": "/" might also work if you manually set the right PKG_CONFIG_PATH, e.g. /usr/lib64/pkgconfig
    "use_sysroot": False,
    "use_system_freetype": True,
    "pdf_bundle_freetype": False,
    "use_system_lcms2": True,
    "use_system_libjpeg": True,
    "use_system_libopenjpeg2": True,
    "use_system_libpng": True,
    "use_system_libtiff": True,
    "use_system_zlib": True,
    "use_custom_libcxx": False,
    "use_libcxx_modules": False,
}

IS_ANDROID = Host.system == SysNames.android
if IS_ANDROID:
    DefaultConfig.update({
        "sysroot": str(Host.usr.parent),
        "current_os": "android",
        "target_os": "android",
    })
    del DefaultConfig["use_sysroot"]
    # On Android, it seems that the build system's CPU type statically defaults to "arm", but we want this script to be host-adaptive (plus, "arm64" is the more likely candidate).
    # TODO(future) refactor platform constants from base.py so we can access abstracted OS/CPU separately through sub-attributes
    AndroidCPUMap = {
        "aarch64": "arm64",
        "armv7l":  "arm",
        "x86_64":  "x64",
        "i686":    "x86",
    }
    raw_cpu = Host._raw_machine
    if raw_cpu in AndroidCPUMap:
        cpu = AndroidCPUMap[raw_cpu]
        DefaultConfig.update({
            "current_cpu": cpu,
            "target_cpu": cpu,
        })
    else:
        log(f"Warning: Unknown Android CPU {raw_cpu}")


def _get_repo(url, target_dir, rev, reset=False, depth=1):
    
    if target_dir.exists():
        if reset:
            log(f"Resetting {target_dir.name} as per --reset option.")
            run_cmd(["git", "reset", "--hard"], cwd=target_dir)
            return True
        else:
            return False
    
    if callable(rev):
        rev = rev()  # resolve deferred
    
    # https://stackoverflow.com/questions/31278902/how-to-shallow-clone-a-specific-commit-with-depth-1
    mkdir(target_dir)
    run_cmd(["git", "init"], cwd=target_dir)
    run_cmd(["git", "remote", "add", "origin", url], cwd=target_dir)
    run_cmd(["git", "fetch", "--depth", str(depth), "origin", rev], cwd=target_dir)
    run_cmd(["git", "checkout", "FETCH_HEAD"], cwd=target_dir)
    
    return True


DEPS_RE = r"\s*'{key}': '(\w+)'"

class _DeferredInfo:
    
    def __init__(self, deps_fields):
        self.deps_fields = deps_fields
    
    @cached_property  # included from base.py
    def deps(self):
        # TODO get a proper parser for the DEPS file format?
        deps_content = (PDFIUM_DIR/"DEPS").read_text()
        result = {}
        for field in self.deps_fields:
            field_re = DEPS_RE.format(key=f"{field}_revision")
            match = re.search(field_re, deps_content)
            assert match, f"Could not find {field!r} in DEPS file"
            result[field] = match.group(1)
        log(f"Found DEPS revisions:\n{result}")
        return result


def _fetch_dep(info, name, target_dir, reset=False):
    # parse out DEPS revisions only when we actually need them
    return _get_repo(DEPS_URLS[name], target_dir, rev=lambda: info.deps[name], reset=reset)


def autopatch(file, pattern, repl, is_regex, exp_count=None):
    log(f"Patch {pattern!r} -> {repl!r} (is_regex={is_regex}) on {file}")
    content = file.read_text()
    if is_regex:
        content, n_subs = re.subn(pattern, repl, content)
    else:
        n_subs = content.count(pattern)
        content = content.replace(pattern, repl)
    if exp_count is not None:
        assert n_subs == exp_count
    file.write_text(content)

def autopatch_dir(dir, globexpr, pattern, repl, is_regex, exp_count=None):
    for file in dir.glob(globexpr):
        autopatch(file, pattern, repl, is_regex, exp_count)


def get_sources(deps_info, short_ver, with_tests, compiler, clang_path, reset, vendor_deps):
    
    assert not IGNORE_FULLVER
    full_ver, pdfium_rev, chromium_rev = handle_sbuild_vers(short_ver)
    
    # pass through reset only for the repositories we actually patch
    do_patches = _get_repo(PDFIUM_URL, PDFIUM_DIR, rev=pdfium_rev, reset=reset)
    if do_patches:
        autopatch_dir(
            PDFIUM_DIR/"public"/"cpp", "*.h",
            r'"public/(.+)"', r'"../\1"',
            is_regex=True, exp_count=None,
        )
        # don't build the test fonts (needed for embedder tests only)
        autopatch(
            PDFIUM_DIR/"testing"/"BUILD.gn",
            r'(\s*)("//third_party/test_fonts")', r"\1# \2",
            is_regex=True, exp_count=1,
        )
        # bundle dependencies (e.g. abseil) into the pdfium DLL
        autopatch(
            PDFIUM_DIR/"BUILD.gn",
            'component("pdfium")',
            'shared_library("pdfium")',
            is_regex=False, exp_count=1,
        )
        autopatch(
            PDFIUM_DIR/"public"/"fpdfview.h",
            "#if defined(COMPONENT_BUILD)",
            "#if 1  // defined(COMPONENT_BUILD)",
            is_regex=False, exp_count=1,
        )
        if sys.byteorder == "big":
            git_apply_patch(PatchDir/"bigendian.patch", cwd=PDFIUM_DIR)
            if with_tests:
                git_apply_patch(PatchDir/"bigendian_test.patch", cwd=PDFIUM_DIR)
    
    do_patches = _fetch_dep(deps_info, "build", PDFIUM_DIR/"build", reset=reset)
    if do_patches:
        # siso.patch: Work around error about path_exists() being undefined.
        # This happens with older versions of GN.
        # Recent GN binaries can be obtained from https://chrome-infra-packages.appspot.com/p/gn/gn
        # Note that merely calling depot_tools `gn` is not sufficient, as it is only a wrapper script looking for vendored GN in the target repository, and if not present (as in this case), falls back to system GN.
        git_apply_patch(PatchDir/"siso.patch", cwd=PDFIUM_DIR/"build")
        if IS_ANDROID:
            # fix linkage step
            git_apply_patch(PatchDir/"android_build.patch", cwd=PDFIUM_DIR/"build")
        if IS_CIBUILDWHEEL:
            # compatibility patch for older system libraries from container
            git_apply_patch(PatchDir/"cibuildwheel.patch", cwd=PDFIUM_DIR)
        if compiler is Compiler.gcc:
            # https://crbug.com/402282789
            git_apply_patch(PatchDir/"ffp_contract.patch", cwd=PDFIUM_DIR/"build")
        elif compiler is Compiler.clang:
            # https://crbug.com/410883044
            clang_patches = ("system_libcxx_with_clang", "avoid_new_clang_flags")
            for patchname in clang_patches:
                git_apply_patch(PatchDir/f"{patchname}.patch", cwd=PDFIUM_DIR/"build")
            # TODO should we handle other OSes here?
            # see also https://groups.google.com/g/llvm-dev/c/k3q_ATl-K_0/m/MjEb6gsCCAAJ
            lld_path = clang_path/"bin"/"ld.lld"
            autopatch(
                PDFIUM_DIR/"build"/"config"/"compiler"/"BUILD.gn",
                'ldflags += [ "-fuse-ld=lld" ]',
                f'ldflags += [ "-fuse-ld={lld_path}" ]',
                is_regex=False, exp_count=1,
            )
    
    do_patches = _fetch_dep(deps_info, "abseil", PDFIUM_3RDPARTY/"abseil-cpp", reset=reset)
    if do_patches and (Host._raw_machine, Host._libc_name) == ("ppc64le", "musl"):
        git_apply_patch(PatchDir/"abseil_ppc64le_musl.patch", cwd=PDFIUM_3RDPARTY/"abseil-cpp")
    
    _fetch_dep(deps_info, "fast_float", PDFIUM_3RDPARTY/"fast_float"/"src")
    if IS_ANDROID:
        _fetch_dep(deps_info, "catapult", PDFIUM_3RDPARTY/"catapult")
    if "icu" in vendor_deps:
        _fetch_dep(deps_info, "icu", PDFIUM_3RDPARTY/"icu")
    if with_tests:
        _fetch_dep(deps_info, "gtest", PDFIUM_3RDPARTY/"googletest"/"src")
        _fetch_dep(deps_info, "test_fonts", PDFIUM_3RDPARTY/"test_fonts")
    
    get_shimheaders_tool(PDFIUM_DIR, rev=chromium_rev)
    
    return full_ver


def prepare(config_dict, build_dir, vendor_deps):
    # Create an empty gclient config
    (PDFIUM_DIR/"build"/"config"/"gclient_args.gni").touch(exist_ok=True)
    if "icu" not in vendor_deps:
        # Unbundle ICU
        # alternatively, we could call build/linux/unbundle/replace_gn_files.py --system-libraries icu
        (PDFIUM_3RDPARTY/"icu").mkdir(exist_ok=True)
        shutil.copyfile(
            PDFIUM_DIR/"build"/"linux"/"unbundle"/"icu.gn",
            PDFIUM_3RDPARTY/"icu"/"BUILD.gn"
        )
    # Create target dir (or reuse existing) and write build config
    mkdir(build_dir)
    # Remove existing libraries from the build dir, to avoid packing unnecessary DLLs when a single-lib build is done after a separate-libs build. This also ensures we really built a new DLL in the end.
    # Leave the object files in place to reuse as much as possible, though.
    for lib in build_dir.glob(Host.libname_glob):
        lib.unlink()
    config_str = serialize_gn_config(config_dict)
    (build_dir/"args.gn").write_text(config_str)


def build(with_tests, build_dir, n_jobs):
    
    ninja_args = []
    if n_jobs is not None:
        ninja_args.extend(["-j", str(n_jobs)])
    
    targets = ["pdfium"]
    if with_tests:
        targets.append("pdfium_unittests")
    
    build_dir_rel = build_dir.relative_to(PDFIUM_DIR)
    run_cmd(["gn", "gen", str(build_dir_rel)], cwd=PDFIUM_DIR)
    run_cmd(["ninja", *ninja_args, "-C", str(build_dir_rel), *targets], cwd=PDFIUM_DIR)


def test(build_dir):
    # FlateModule.Encode may fail with older zlib (generates different results)
    os.environ["GTEST_FILTER"] = "*-FlateModule.Encode"
    run_cmd([build_dir/"pdfium_unittests"], cwd=PDFIUM_DIR, check=False)


def _get_clang_ver(clang_path):
    from packaging.version import Version
    output = run_cmd([str(clang_path/"bin"/"clang"), "--version"], capture=True, cwd=None)
    log(output)
    version = re.search(r"version ([\d\.]+)", output).group(1)
    version = Version(version).major
    log(f"Determined clang version {version!r}")
    return version

def setup_compiler(config, compiler, clang_path):
    if compiler is Compiler.gcc:
        config["is_clang"] = False
    elif compiler is Compiler.clang:
        assert clang_path, "Clang path must be set"
        clang_version = _get_clang_ver(clang_path)
        config.update({
            "is_clang": True,
            "clang_base_path": str(clang_path),
            "clang_version": clang_version,
        })
    else:
        assert False, f"Unhandled compiler {compiler}"


def main(build_ver=None, with_tests=False, n_jobs=None, compiler=None, clang_path=None, reset=False, vendor_deps=None):
    
    if build_ver is None:
        build_ver = SBUILD_NATIVE_PIN
    if vendor_deps is None:
        vendor_deps = set()
    if compiler is None:
        if shutil.which("gcc"):
            compiler = Compiler.gcc
        elif shutil.which("clang"):
            log("gcc not available, will try clang. Note, you may need to set up some symlinks to match the clang directory layout expected by pdfium. Also, make sure libclang_rt builtins are installed.")
            compiler = Compiler.clang
        else:
            raise RuntimeError("Neither gcc nor clang installed.")
    if compiler is Compiler.clang and clang_path is None:
        clang_path = Host.usr
    
    build_dir = PDFIUM_DIR/"out"/"Default"
    config = DefaultConfig.copy()
    
    deps_fields = ["build", "abseil", "fast_float"]
    if IS_ANDROID:
        deps_fields.append("catapult")
    if "icu" in vendor_deps:
        deps_fields.append("icu")
    if with_tests:
        deps_fields += ("gtest", "test_fonts")
    
    deps_info = _DeferredInfo(deps_fields)
    
    mkdir(SOURCES_DIR)
    full_ver = get_sources(deps_info, build_ver, with_tests, compiler, clang_path, reset, vendor_deps)
    setup_compiler(config, compiler, clang_path)
    prepare(config, build_dir, vendor_deps)
    build(with_tests, build_dir, n_jobs)
    if with_tests:
        test(build_dir)
    
    return pack_sourcebuild(PDFIUM_DIR, build_dir, "native", full_ver, build_ver)


def parse_args(argv):
    parser = argparse.ArgumentParser(
        description = "Build PDFium from source natively with system tools/libraries. This does not use Google's binary toolchain, so it should be portable across different Linux architectures. Whether this might also work on other OSes depends on PDFium's build system and the availability of a Linux-like system library environment.",
    )
    parser.add_argument(
        "--version",
        dest = "build_ver",
        help = f"The pdfium version to use. Currently defaults to {SBUILD_NATIVE_PIN}. Pass 'main' to try the latest state.",
    )
    parser.add_argument(
        "--test",
        dest = "with_tests",
        action = "store_true",
        help = "Whether to build and run tests. Recommended, except on very slow hosts.",
    )
    parser.add_argument(
        "-j", "--jobs",
        dest = "n_jobs",
        type = int,
        metavar = "N",
        help = "The number of build jobs to use. If not given, ninja will choose this value. Pass -j $(nproc) if you wanna make sure this matches the number of processor cores.",
    )
    parser.add_argument(
        "-c", "--compiler",
        type = str.lower,
        help = "The compiler to use (gcc or clang). Defaults to gcc if available.",
    )
    parser.add_argument(
        "--reset",
        action = "store_true",
        help = "Reset those git repos that we patch, and re-apply the patches. This is necessary when making a rebuild with different patch configuration (e.g. when switching between gcc <-> clang), but is not enabled by default to avoid unintentional loss of manual changes.",
    )
    # Hint: If you have a simultaneous toolchained checkout, you could use e.g. './sbuild/toolchained/pdfium/third_party/llvm-build/Release+Asserts'
    parser.add_argument(
        "--clang-path",
        type = lambda p: Path(p).expanduser().resolve(),
        help = "Path to clang release folder, without trailing slash. Passing `--compiler clang` is a pre-requisite. By default, we try '/usr' or similar, but your system's folder structure might not match the layout expected by pdfium. Consider creating symlinks as described in pypdfium2's README.md.",
    )
    # The --vendor option is provided for cibuildwheel clients:
    # - libicudata pulled in from the system via `auditwheel repair` is quite big. Using vendored ICU reduces wheel size by about 10 MB (compressed).
    # - libc++ is used but not pulled in by auditwheel. This appears to be ABI-unsafe (although the wheels seem to work across different hosts according to downstream feedback), so we may want to add that in the future. Actually, options to use system libc++ are deprecated upstream anyway.
    parser.add_argument(
        "--vendor",
        dest = "vendor_deps",
        nargs = "+",
        action = "extend",
        help = "Dependencies to vendor. Note, this only supports libraries where there is a specific reason to vendor despite the native build. Currently this means 'icu' only ('libc++' may be added in the future). For an exhaustive vendored build, use build_toolchained.py"
    )
    args = parser.parse_args(argv)
    if args.compiler:
        args.compiler = Compiler[args.compiler]
    if args.vendor_deps:
        args.vendor_deps = set(args.vendor_deps)
    return args


def main_cli():
    args = parse_args(sys.argv[1:])
    main(**vars(args))


if __name__ == "__main__":
    main_cli()



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# platform files are managed in an own pypdfium2_raw module, cleanly separating raw bindings and helpers
from pypdfium2_raw.bindings import *  # noqa



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pypdfium2.raw as pdfium_c
from pypdfium2.version import PDFIUM_INFO


class _fallback_dict (dict):
    
    def get(self, key, default_prefix="Unhandled constant"):
        return dict.get(self, key, f"{default_prefix} {key}")


#: Convert a rotation value in degrees to a PDFium constant.
RotationToConst = {
    0:   0,
    90:  1,
    180: 2,
    270: 3,
}

#: Convert a PDFium rotation constant to a value in degrees. Inversion of :data:`.RotationToConst`.
RotationToDegrees = {v: k for k, v in RotationToConst.items()}


#: Get the number of channels for a PDFium bitmap format. (:attr:`FPDFBitmap_Unknown` is deliberately not handled.)
BitmapTypeToNChannels = {
    pdfium_c.FPDFBitmap_Gray: 1,
    pdfium_c.FPDFBitmap_BGR:  3,
    pdfium_c.FPDFBitmap_BGRx: 4,
    pdfium_c.FPDFBitmap_BGRA: 4,
}

#: Convert a PDFium bitmap format to string, assuming BGR byte order. (:attr:`FPDFBitmap_Unknown` is deliberately not handled.)
BitmapTypeToStr = {
    pdfium_c.FPDFBitmap_Gray: "L",
    pdfium_c.FPDFBitmap_BGR:  "BGR",
    pdfium_c.FPDFBitmap_BGRx: "BGRX",
    pdfium_c.FPDFBitmap_BGRA: "BGRA",
}

#: Convert a PDFium bitmap format to string, assuming RGB byte order. (:attr:`FPDFBitmap_Unknown` is deliberately not handled.)
BitmapTypeToStrReverse = {
    pdfium_c.FPDFBitmap_Gray: "L",
    pdfium_c.FPDFBitmap_BGR:  "RGB",
    pdfium_c.FPDFBitmap_BGRx: "RGBX",
    pdfium_c.FPDFBitmap_BGRA: "RGBA",
}

if PDFIUM_INFO.build >= 7098:
    # New pixel format FPDFBitmap_BGRA_Premul
    # Skia-only at the time of writing. Added for completeness and to satisfy the test suite.
    BitmapTypeToNChannels[pdfium_c.FPDFBitmap_BGRA_Premul] = 4
    BitmapTypeToStr[pdfium_c.FPDFBitmap_BGRA_Premul] = "BGRa"
    BitmapTypeToStrReverse[pdfium_c.FPDFBitmap_BGRA_Premul] = "RGBa"

# TODO consider a bi-directional dict in the future?
#: Convert a string to PDFium bitmap format, assuming BGR byte order. Inversion of :data:`BitmapTypeToStr`.
BitmapStrToConst = {v: k for k, v in BitmapTypeToStr.items()}

#: Convert a string to PDFium bitmap format, assuming RGB byte order. Inversion of :data:`BitmapTypeToStrReverse`.
BitmapStrReverseToConst = {v: k for k, v in BitmapTypeToStrReverse.items()}

#: Convert a PDFium form type (:attr:`FORMTYPE_*`) to string.
FormTypeToStr = _fallback_dict({
    pdfium_c.FORMTYPE_NONE:           "None",
    pdfium_c.FORMTYPE_ACRO_FORM:      "AcroForm",
    pdfium_c.FORMTYPE_XFA_FULL:       "XFA",
    pdfium_c.FORMTYPE_XFA_FOREGROUND: "XFAF",
})

#: Convert a PDFium color space constant (:attr:`FPDF_COLORSPACE_*`) to string.
ColorspaceToStr = _fallback_dict({
    pdfium_c.FPDF_COLORSPACE_UNKNOWN:    "?",
    pdfium_c.FPDF_COLORSPACE_DEVICEGRAY: "DeviceGray",
    pdfium_c.FPDF_COLORSPACE_DEVICERGB:  "DeviceRGB",
    pdfium_c.FPDF_COLORSPACE_DEVICECMYK: "DeviceCMYK",
    pdfium_c.FPDF_COLORSPACE_CALGRAY:    "CalGray",
    pdfium_c.FPDF_COLORSPACE_CALRGB:     "CalRGB",
    pdfium_c.FPDF_COLORSPACE_LAB:        "Lab",
    pdfium_c.FPDF_COLORSPACE_ICCBASED:   "ICCBased",
    pdfium_c.FPDF_COLORSPACE_SEPARATION: "Separation",
    pdfium_c.FPDF_COLORSPACE_DEVICEN:    "DeviceN",
    pdfium_c.FPDF_COLORSPACE_INDEXED:    "Indexed",  # i.e. palettized
    pdfium_c.FPDF_COLORSPACE_PATTERN:    "Pattern",
})

#: Convert a PDFium view mode constant (:attr:`PDFDEST_VIEW_*`) to string.
ViewmodeToStr = _fallback_dict({
    pdfium_c.PDFDEST_VIEW_UNKNOWN_MODE: "?",
    pdfium_c.PDFDEST_VIEW_XYZ:   "XYZ",
    pdfium_c.PDFDEST_VIEW_FIT:   "Fit",
    pdfium_c.PDFDEST_VIEW_FITH:  "FitH",
    pdfium_c.PDFDEST_VIEW_FITV:  "FitV",
    pdfium_c.PDFDEST_VIEW_FITR:  "FitR",
    pdfium_c.PDFDEST_VIEW_FITB:  "FitB",
    pdfium_c.PDFDEST_VIEW_FITBH: "FitBH",
    pdfium_c.PDFDEST_VIEW_FITBV: "FitBV",
})

#: Convert a PDFium object type constant (:attr:`FPDF_PAGEOBJ_*`) to string.
ObjectTypeToStr = _fallback_dict({
    pdfium_c.FPDF_PAGEOBJ_UNKNOWN: "?",
    pdfium_c.FPDF_PAGEOBJ_TEXT:    "text",
    pdfium_c.FPDF_PAGEOBJ_PATH:    "path",
    pdfium_c.FPDF_PAGEOBJ_IMAGE:   "image",
    pdfium_c.FPDF_PAGEOBJ_SHADING: "shading",
    pdfium_c.FPDF_PAGEOBJ_FORM:    "form",
})

#: Convert an object type string to a PDFium constant. Inversion of :data:`.ObjectTypeToStr`.
ObjectTypeToConst = {v: k for k, v in ObjectTypeToStr.items()}

#: Convert a PDFium page mode constant (:attr:`PAGEMODE_*`) to string.
PageModeToStr = _fallback_dict({
    pdfium_c.PAGEMODE_UNKNOWN:        "?",
    pdfium_c.PAGEMODE_USENONE:        "None",
    pdfium_c.PAGEMODE_USEOUTLINES:    "Outline",
    pdfium_c.PAGEMODE_USETHUMBS:      "Thumbnails",
    pdfium_c.PAGEMODE_FULLSCREEN:     "Full-screen",
    pdfium_c.PAGEMODE_USEOC:          "Layers",
    pdfium_c.PAGEMODE_USEATTACHMENTS: "Attachments",
})

#: Convert a PDFium error constant (:attr:`FPDF_ERR_*`) to string.
ErrorToStr = _fallback_dict({
    pdfium_c.FPDF_ERR_SUCCESS:  "Success",
    pdfium_c.FPDF_ERR_UNKNOWN:  "Unknown error",
    pdfium_c.FPDF_ERR_FILE:     "File access error",
    pdfium_c.FPDF_ERR_FORMAT:   "Data format error",
    pdfium_c.FPDF_ERR_PASSWORD: "Incorrect password error",
    pdfium_c.FPDF_ERR_SECURITY: "Unsupported security scheme error",
    pdfium_c.FPDF_ERR_PAGE:     "Page not found or content error",
})


if "XFA" in PDFIUM_INFO.flags:  # pragma: no cover
    #: [XFA builds only] Convert a PDFium XFA error constant (:attr:`FPDF_ERR_XFA*`) to string.
    XFAErrorToStr = _fallback_dict({
        pdfium_c.FPDF_ERR_XFALOAD:   "Load error",
        pdfium_c.FPDF_ERR_XFALAYOUT: "Layout error",
    })

#: Convert a PDFium unsupported constant (:attr:`FPDF_UNSP_*`) to string.
UnsupportedInfoToStr = _fallback_dict({
    pdfium_c.FPDF_UNSP_DOC_XFAFORM:               "XFA form",
    pdfium_c.FPDF_UNSP_DOC_PORTABLECOLLECTION:    "Portable collection",
    # https://crbug.com/pdfium/1945
    pdfium_c.FPDF_UNSP_DOC_ATTACHMENT:            "Attachment (incomplete support)",
    pdfium_c.FPDF_UNSP_DOC_SECURITY:              "Security",
    pdfium_c.FPDF_UNSP_DOC_SHAREDREVIEW:          "Shared review",
    pdfium_c.FPDF_UNSP_DOC_SHAREDFORM_ACROBAT:    "Shared form (acrobat)",
    pdfium_c.FPDF_UNSP_DOC_SHAREDFORM_FILESYSTEM: "Shared form (filesystem)",
    pdfium_c.FPDF_UNSP_DOC_SHAREDFORM_EMAIL:      "Shared form (email)",
    pdfium_c.FPDF_UNSP_ANNOT_3DANNOT:             "3D annotation",
    pdfium_c.FPDF_UNSP_ANNOT_MOVIE:               "Movie annotation",
    pdfium_c.FPDF_UNSP_ANNOT_SOUND:               "Sound annotation",
    pdfium_c.FPDF_UNSP_ANNOT_SCREEN_MEDIA:        "Screen media annotation",
    pdfium_c.FPDF_UNSP_ANNOT_SCREEN_RICHMEDIA:    "Screen rich media annotation",
    pdfium_c.FPDF_UNSP_ANNOT_ATTACHMENT:          "Attachment annotation",
    pdfium_c.FPDF_UNSP_ANNOT_SIG:                 "Signature annotation",
})



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# NOTE: unless otherwise noted, "version" refers to the short version in this file

import sys
import shutil
import tarfile
import argparse
import functools
from pathlib import Path
import urllib.request as url_request
from concurrent.futures import ThreadPoolExecutor

# local
from base import *
from _verify import do_verify


def urlretrieve(url, fp, *args, **kwargs):
    log(f"{url!r} -> {str(fp)!r}")
    url_request.urlretrieve(url, fp, *args, **kwargs)


def clear_data(download_files):
    for pl_name in download_files:
        pl_dir = DataDir / pl_name
        if pl_dir.exists():
            shutil.rmtree(pl_dir)


def _get_package(pl_name, version, robust, use_v8):
    
    pl_dir = DataDir / pl_name
    mkdir(pl_dir)
    
    prefix = "pdfium-"
    if use_v8:
        prefix += "v8-"
    
    fn = prefix + f"{PdfiumBinariesMap[pl_name]}.tgz"
    fu = f"{ReleaseURL}{version}/{fn}"
    fp = pl_dir / fn
    
    try:
        urlretrieve(fu, fp)
    except Exception as e:
        if robust:
            log(str(e))
            return None, None
        else:
            raise
    
    return pl_name, fp


def do_download(platforms, version, use_v8, max_workers, robust):
    
    if not max_workers:
        max_workers = len(platforms)
    
    archives = {}
    with ThreadPoolExecutor(max_workers=max_workers) as pool:
        func = functools.partial(_get_package, version=version, robust=robust, use_v8=use_v8)
        for pl_name, file_path in pool.map(func, platforms):
            if pl_name is not None:
                archives[pl_name] = file_path
    
    return archives


def _parse_ver_file(buffer, short_ver):
    content = buffer.read().decode().strip()
    full_ver = [int(l.split("=")[-1].strip()) for l in content.split("\n")]
    full_ver = PdfiumVer.scheme(*full_ver)
    assert full_ver.build == short_ver
    if full_ver in PdfiumVer._vdict:
        assert PdfiumVer._vdict[short_ver] == full_ver
    else:
        PdfiumVer._vdict[full_ver.build] = full_ver
    log(f"Resolved {short_ver} -> {full_ver} (by pdfium-binaries VERSION)")
    return full_ver


def _extract_licenses(tar, pl_dir):
    licenses_dir = pl_dir/"BUILD_LICENSES"
    mkdir(licenses_dir)
    all_paths = tar.getnames()
    if "licenses" in all_paths:
        license_paths = (p for p in all_paths if p.startswith("licenses/"))
        for path in license_paths:
            tar_extract_file(tar, path, licenses_dir/Path(path).name)
        tar_extract_file(tar, "LICENSE", licenses_dir/"pdfium-binaries.txt")
    else:
        # backwards compat with pdfium-binaries < 7175
        tar_extract_file(tar, "LICENSE", licenses_dir/"aggregated-license.txt")


def do_extract(archives, version, flags):
    
    for pl_name, arc_path in archives.items():
        
        with tarfile.open(arc_path) as tar:
            pl_dir = DataDir/pl_name
            system = plat_to_system(pl_name)
            libname = libname_for_system(system)
            tar_libdir = "lib" if system != SysNames.windows else "bin"
            tar_extract_file(tar, f"{tar_libdir}/{libname}", pl_dir/libname)
            _extract_licenses(tar, pl_dir)
            full_ver = _parse_ver_file(tar.extractfile("VERSION"), version)
            write_pdfium_info(pl_dir, full_ver, origin="pdfium-binaries", flags=flags)
        
        arc_path.unlink()


def postprocess_android():
    # see https://wiki.termux.com/wiki/FAQ#Why_does_a_compiled_program_show_warnings
    elf_cleaner = shutil.which("termux-elf-cleaner")
    if elf_cleaner:
        log("Invoking termux-elf-cleaner to clean up possible linker warnings...")
        libpath = DataDir / Host.platform / libname_for_system(Host.system)
        run_cmd([elf_cleaner, str(libpath)], cwd=None)
    else:
        log("If you are on Termux, consider installing termux-elf-cleaner to clean up possible linker warnings.")


def main(platforms, version, robust=False, max_workers=None, use_v8=False, verify=None):
    
    if not platforms:
        platforms = WheelPlatforms
    if len(platforms) != len(set(platforms)):
        raise ValueError("Duplicate platforms not allowed.")
    flags = ("V8", "XFA") if use_v8 else ()
    
    clear_data(platforms)
    archives = do_download(platforms, version, use_v8, max_workers, robust)
    do_verify(verify, archives, version)
    
    do_extract(archives, version, flags)
    if Host.system == SysNames.android and Host.platform in platforms:
        postprocess_android()


# low-level interface for internal use - end users should go with cached, higher-level emplace.py or setup.py instead

def parse_args(argv):
    platform_choices = list(PdfiumBinariesMap.keys())
    parser = argparse.ArgumentParser(
        description = "Download pre-built PDFium packages.",
    )
    parser.add_argument(
        "--platforms", "-p",
        nargs = "+",
        metavar = "ID",
        choices = platform_choices,
        help = f"The platform(s) to include. Defaults to the platforms we build wheels for. Choices: {platform_choices}",
    )
    parser.add_argument(
        "--use-v8",
        action = "store_true",
        help = "Use V8 binaries (JavaScript/XFA support)."
    )
    parser.add_argument(
        "--version", "-v",
        help = "The binaries release to use. Either 'latest' (the default), 'pinned', or a pdfium-binaries tag integer."
    )
    parser.add_argument(
        "--max-workers",
        type = int,
        help = "Maximum number of jobs to run in parallel when downloading binaries.",
    )
    parser.add_argument(
        "--robust",
        action = "store_true",
        help = "Skip missing binaries instead of raising an exception.",
    )
    parser.add_argument(
        "--verify",
        action = "store_true",
        default = None,
        help = "Verify release artifacts through GitHub build provenance attestations. This will be automatically enabled if `gh` is installed and the requested pdfium version is recent enough.",
    )
    return parser.parse_args(argv)


def cli_main(argv=sys.argv[1:]):
    args = parse_args(argv)
    if not args.version or args.version == "latest":
        args.version = PdfiumVer.get_latest()
    elif args.version == "pinned":
        args.version = PdfiumVer.pinned
    else:
        args.version = int(args.version)
    main(**vars(args))


if __name__ == "__main__":
    cli_main()



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# See also https://stackoverflow.com/questions/45150304/how-to-force-a-python-wheel-to-be-platform-specific-when-building-it and https://github.com/innodatalabs/redstork/blob/master/setup.py

import os
import sys
from pathlib import Path
import setuptools
from setuptools.command.build_py import build_py as build_py_orig
try:
    from setuptools.command.bdist_wheel import bdist_wheel
except ImportError:
    from wheel.bdist_wheel import bdist_wheel

sys.path.insert(0, str(Path(__file__).parent/"setupsrc"))
from base import *
from emplace import prepare_setup


# Use a custom distclass declaring we have a binary extension, to prevent modules from being nested in a purelib/ subdirectory in wheels.
# This will also set `Root-Is-Purelib: false` in the WHEEL file, and make the wheel tag platform specific by default.

class BinaryDistribution (setuptools.Distribution):
    
    def has_ext_modules(self):
        return True


def bdist_factory(pl_name):
    
    class pypdfium_bdist (bdist_wheel):
        
        def finalize_options(self, *args, **kws):
            bdist_wheel.finalize_options(self, *args, **kws)
            # should be handled by the distclass already, but set it again to be on the safe side
            self.root_is_pure = False
        
        def get_tag(self, *args, **kws):
            if pl_name == ExtPlats.sourcebuild:
                # In case of cross-compilation (or even just proper packaging), the caller needs to set the tag.
                plat_tag = os.environ.get("CROSS_TAG")
                # Otherwise, forward the host's tag as given by bdist_wheel.
                # Alternatively, the sourcebuild clause in base.py::get_wheel_tag() should be roughly equivalent.
                # The underlying API being wrapped is sysconfig.get_platform().
                if not plat_tag:
                    _py, _abi, plat_tag = bdist_wheel.get_tag(self, *args, **kws)
            else:
                plat_tag = get_wheel_tag(pl_name)
            return "py3", "none", plat_tag
    
    return pypdfium_bdist


class pypdfium_build_py (build_py_orig):
        
    def run(self, *args, **kwargs):
        if hasattr(self, "editable_mode"):
            helpers_info = read_json(ModuleDir_Helpers/VersionFN)
            helpers_info["is_editable"] = bool(self.editable_mode)
            write_json(ModuleDir_Helpers/VersionFN, helpers_info)
        else:
            log("!!! Warning: cmdclass does not provide `editable_mode` attribute.")
        build_py_orig.run(self, *args, **kwargs)


LICENSES_SHARED = (
    "LICENSES/Apache-2.0.txt",
    "LICENSES/BSD-3-Clause.txt",
    "LICENSES/CC-BY-4.0.txt",
)
LICENSES_SDIST = (
    # our sdists don't currently include tests, so we don't need to list the other licenses here
    "REUSE.toml",
)


def assert_exists(dir, data_files):
    missing = tuple(f for f in data_files if not (dir/f).exists())
    if missing:
        assert False, f"Missing data files: {missing}"


def run_setup(modnames, pl_name, platfiles):
    
    # FIXME ambiguity between `pl_name == ExtPlats.sdist` and `ModuleRaw not in modnames` ?
    
    kwargs = dict(
        name = "pypdfium2",
        description = "Python bindings to PDFium",
        license = "BSD-3-Clause, Apache-2.0, dependency licenses",
        python_requires = ">= 3.6",
        cmdclass = {},
        package_dir = {},
        package_data = {},
        install_requires = [],
    )
    
    license_files = list(LICENSES_SHARED)
    if pl_name == ExtPlats.sdist:
        license_files.extend(LICENSES_SDIST)
    
    if modnames == [ModuleHelpers]:
        kwargs["name"] += "_helpers"
        kwargs["description"] += " (helpers module)"
        kwargs["install_requires"].append("pypdfium2_raw")
    elif modnames == [ModuleRaw]:
        kwargs["name"] += "_raw"
        kwargs["description"] += " (raw module)"
        pdfium_fullver, _ = read_pdfium_info(ModuleDir_Raw)
        kwargs["version"] = str(pdfium_fullver)
    else:
        assert any(m in modnames for m in (ModuleHelpers, ModuleRaw)), \
               f"At least one core module is required. Check {ModulesSpec_EnvVar}."
    
    if ModuleHelpers in modnames:
        helpers_info = get_helpers_info()
        if pl_name == ExtPlats.sdist:
            if helpers_info["dirty"]:
                # ignore dirty state due to craft.py::tmp_ctypesgen_pin()
                if int(os.environ.get("SDIST_IGNORE_DIRTY", 0)):
                    helpers_info["dirty"] = False
            else:
                log("Warning: sdist without ctypesgen pin, or git describe not working?")
        kwargs["version"] = merge_tag(helpers_info, mode="py")
        # is_editable = None: unknown/fallback in case the cmdclass is not reached
        helpers_info["is_editable"] = None
        write_json(ModuleDir_Helpers/VersionFN, helpers_info)
        kwargs["cmdclass"]["build_py"] = pypdfium_build_py
        kwargs["package_dir"]["pypdfium2"] = "src/pypdfium2"
        kwargs["package_data"]["pypdfium2"] = (VersionFN, )
        kwargs["entry_points"] = dict(console_scripts=["pypdfium2 = pypdfium2.__main__:cli_main"])
    if ModuleRaw in modnames:
        kwargs["package_dir"]["pypdfium2_raw"] = "src/pypdfium2_raw"
        if platfiles:
            kwargs["package_data"]["pypdfium2_raw"] = platfiles
    
    if ModuleRaw not in modnames or pl_name == ExtPlats.sdist:
        kwargs["exclude_package_data"] = {"pypdfium2_raw": (VersionFN, BindingsFN, *LIBNAME_GLOBS)}
    elif pl_name == ExtPlats.system:
        kwargs["exclude_package_data"] = {"pypdfium2_raw": LIBNAME_GLOBS}
    else:
        if pl_name == ExtPlats.sourcebuild:
            license_files.append("BUILD_LICENSES/*")
        else:
            # FIXME This gives a deeply nested directory structure.
            # The author is not aware of a way to achieve a more flat structure with setuptools.
            license_files.append(f"data/{pl_name}/BUILD_LICENSES/*")
        kwargs["distclass"] = BinaryDistribution
        kwargs["cmdclass"]["bdist_wheel"] = bdist_factory(pl_name)
    
    kwargs["license_files"] = license_files
    
    # check
    if "pypdfium2" in kwargs["package_data"]:
        assert_exists(ModuleDir_Helpers, kwargs["package_data"]["pypdfium2"])
    if "pypdfium2_raw" in kwargs["package_data"]:
        assert_exists(ModuleDir_Raw, kwargs["package_data"]["pypdfium2_raw"])
    
    setuptools.setup(**kwargs)


def main():
    
    raw_modspec = os.environ.get(ModulesSpec_EnvVar, "")
    raw_platspec = os.environ.get(PlatSpec_EnvVar, "")
    
    modnames = parse_modspec(raw_modspec)
    pl_name, sub_target, requested_ver, flags = parse_pl_spec(raw_platspec)
    
    if pl_name == ExtPlats.sdist and modnames != ModulesAll:
        raise ValueError(f"Partial sdist does not make sense - unset {ModulesSpec_EnvVar}.")
    
    if ModuleRaw in modnames and pl_name != ExtPlats.sdist:
        platfiles, pl_name = prepare_setup(pl_name, sub_target, requested_ver, flags)
    else:
        platfiles = ()
    
    run_setup(modnames, pl_name, platfiles)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# TODO(future) add bindings info (e.g. ctypesgen version, reference/generated, runtime libpaths)

__all__ = ("PYPDFIUM_INFO", "PDFIUM_INFO")

import json
from pathlib import Path
import pypdfium2_raw


class _version_class:
    
    def __init__(self):
        with open(self._FILE, "r") as buf:
            data = json.load(buf)
        for k, v in data.items():
            setattr(self, k, v)
        self.api_tag = tuple(data[k] for k in self._TAG_FIELDS)
        self._hook()
        self.version = self.tag + self.desc
    
    def __repr__(self):
        return self.version
    
    def _craft_tag(self):
        return ".".join(str(v) for v in self.api_tag)
    
    def _craft_desc(self, *suffixes):
        
        local_ver = []
        if self.n_commits > 0:
            local_ver += [str(self.n_commits), str(self.hash)]
        local_ver += suffixes
        
        desc = ""
        if local_ver:
            desc += "+" + ".".join(local_ver)
        return desc


class _version_pypdfium2 (_version_class):
    
    _FILE = Path(__file__).parent / "version.json"
    _TAG_FIELDS = ("major", "minor", "patch")
    
    def _hook(self):
        
        self.tag = self._craft_tag()
        if self.beta is not None:
            self.tag += f"b{self.beta}"
        
        suffixes = ["dirty"] if self.dirty else []
        self.desc = self._craft_desc(*suffixes)
        if self.data_source != "git":
            self.desc += f":{self.data_source}"
        if self.is_editable:
            self.desc += "@editable"


class _version_pdfium (_version_class):
    
    _FILE = Path(pypdfium2_raw.__file__).parent / "version.json"
    _TAG_FIELDS = ("major", "minor", "build", "patch")
    
    def _hook(self):
        
        self.flags = tuple(self.flags)
        self.tag = self._craft_tag()
        
        self.desc = self._craft_desc()
        if self.flags:
            self.desc += f":{','.join(self.flags)}"
        if self.origin != "pdfium-binaries":
            self.desc += f"@{self.origin}"


PYPDFIUM_INFO = _version_pypdfium2()
"""
pypdfium2 helpers version.

It is suggesed to compare against *api_tag* and possibly also *beta* (see below).

Parameters:
    version (str):
        Joined tag and desc, forming the full version.
    tag (str):
        Version ciphers joined as str, including possible beta. Corresponds to the latest release tag at install time.
    desc (str):
        Non-cipher descriptors represented as str.
    api_tag (tuple[int]):
        Version ciphers joined as tuple, excluding possible beta.
    major (int):
        Major cipher.
    minor (int):
        Minor cipher.
    patch (int):
        Patch cipher.
    beta (int | None):
        Beta cipher, or None if not a beta version.
    n_commits (int):
        Number of commits after tag at install time. 0 for release.
    hash (str | None):
        Hash of head commit if n_commits > 0, None otherwise.
    dirty (bool):
        True if there were uncommitted changes at install time, False otherwise.
    data_source (str):
        Source of this version info. Possible values:\n
        - ``git``: Parsed from git describe. Always used if available. Highest accuracy.
        - ``given``: Pre-supplied version file (e.g. packaged with sdist, or else created by caller).
        - ``record``: Parsed from autorelease record. Implies that possible changes after tag are unknown.
    is_editable (bool | None):
        True for editable install, False otherwise. None if unknown.\n
        If True, the version info is the one captured at install time. An arbitrary number of forward or reverse changes may have happened since.
"""


PDFIUM_INFO = _version_pdfium()
"""
PDFium version.

It is suggesed to compare against *build* (see below).

Parameters:
    version (str):
        Joined tag and desc, forming the full version.
    tag (str):
        Version ciphers joined as string.
    desc (str):
        Descriptors (origin, flags) as string.
    api_tag (tuple[int]):
        Version ciphers grouped as tuple.
    major (int):
        Chromium major cipher.
    minor (int):
        Chromium minor cipher.
    build (int):
        Chromium/pdfium build cipher.
        This value uniquely identifies the pdfium version.
    patch (int):
        Chromium patch cipher.
    n_commits (int):
        Number of commits after tag at install time. 0 for tagged build commit.
    hash (str | None):
        Hash of head commit if n_commits > 0, None otherwise.
    origin (str):
        The pdfium binary's origin.
    flags (tuple[str]):
        Tuple of pdfium feature flags. Empty for default build. (V8, XFA) for pdfium-binaries V8 build.
"""

# Freeze the base class after we have constructed the instance objects
def _frozen_setattr(self, name, value):
    raise AttributeError(f"Version class is read-only - assignment '{name} = {value}' not allowed")
_version_class.__setattr__ = _frozen_setattr



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import argparse
import importlib
from pypdfium2.version import PYPDFIUM_INFO, PDFIUM_INFO
from pypdfium2._cli._parsers import setup_logging

from pypdfium2_raw.bindings import _libs

SubCommands = {
    "arrange":        "Rearrange/merge documents",
    "attachments":    "List/extract/edit embedded files",
    "extract-images": "Extract images",
    "extract-text":   "Extract text",
    "imgtopdf":       "Convert images to PDF",
    "pageobjects":    "Print info on pageobjects",
    "pdfinfo":        "Print info on document and pages",
    "render":         "Rasterize pages",
    "tile":           "Tile pages (N-up)",
    "toc":            "Print table of contents",
}

CmdToModule = {n: importlib.import_module(f"pypdfium2._cli.{n.replace('-', '_')}") for n in SubCommands}


def get_parser():
    
    main_parser = argparse.ArgumentParser(
        prog = "pypdfium2",
        formatter_class = argparse.RawTextHelpFormatter,
        description = "Command line interface to the pypdfium2 library (Python binding to PDFium)",
    )
    main_parser.add_argument(
        "--version", "-v",
        action = "version",
        version = f"pypdfium2 {PYPDFIUM_INFO}\n" f"pdfium {PDFIUM_INFO} at {_libs['pdfium']._name}"
    )
    subparsers = main_parser.add_subparsers(dest="subcommand")
    
    for name, help in SubCommands.items():
        mod = CmdToModule[name]
        desc = getattr(mod, "PARSER_DESC", None)
        desc = (help + "\n\n" + desc) if desc else help
        subparser = subparsers.add_parser(
            name, help=help, description=desc,
            formatter_class=argparse.RawTextHelpFormatter,
        )
        mod.attach(subparser)
    
    return main_parser


def api_main(raw_args=sys.argv[1:]):
    
    parser = get_parser()
    args = parser.parse_args(raw_args)
    
    if not args.subcommand:
        parser.print_help()
        return
    
    CmdToModule[args.subcommand].main(args)


def cli_main():
    setup_logging()
    api_main()


if __name__ == "__main__":
    cli_main()



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# TODO test-cover use_bitmap, format and render

import logging
import traceback
from pathlib import Path
import pypdfium2.raw as pdfium_c
import pypdfium2._helpers as pdfium
from pypdfium2._cli._parsers import (
    add_input, get_input,
    BooleanOptionalAction,
)

logger = logging.getLogger(__name__)


def attach(parser):
    add_input(parser, pages=True)
    parser.add_argument(
        "--output-dir", "-o",
        required = True,
        type = Path,
        help = "Output directory to take the extracted images",
    )
    parser.add_argument(
        "--max-depth",
        type = int,
        default = 15,
        help = "Maximum recursion depth to consider when looking for pageobjects.",
    )
    parser.add_argument(
        "--use-bitmap",
        action = "store_true",
        help = "Enforce the use of bitmaps rather than attempting a smart extraction of the image.",
    )
    parser.add_argument(
        "--format",
        help = "Image format to use when saving bitmaps. (Fallback if doing smart extraction.)",
    )
    parser.add_argument(
        "--render",
        action = "store_true",
        help = "When --use-bitmap is given, whether to get rendered bitmaps, taking masks and transform matrices into account.",
    )
    parser.add_argument(
        "--scale-to-original",
        action = BooleanOptionalAction,
        default = True,
        help = "When --use-bitmap --render is given, whether to scale the image so it is rendered at its native resolution, or close to that. This should improve output quality. The default is True, but you may opt out.",
    )


def main(args):
    
    if not args.output_dir.is_dir():
        raise NotADirectoryError(args.output_dir)
    if args.use_bitmap and not args.format:
        args.format = "png"
    
    pdf = get_input(args)
    n_pdigits = len(str( max(args.pages)+1 ))
    
    for i in args.pages:
        
        page = pdf[i]
        images = page.get_objects(
            filter = (pdfium_c.FPDF_PAGEOBJ_IMAGE, ),
            max_depth = args.max_depth,
        )
        
        # not perfectly memory efficient, but we need image count for digit formatting
        images = list(images)
        n_idigits = len(str( len(images) ))
        
        for j, image in enumerate(images):
            tag = "%0*d_%0*d" % (n_pdigits, i+1, n_idigits, j+1)
            prefix = args.output_dir / f"{args.input.stem}_{tag}"
            # logger.debug("\n"+tag)
            try:
                if args.use_bitmap:
                    pil_image = image.get_bitmap(render=args.render, scale_to_original=args.scale_to_original).to_pil()
                    pil_image.save(f"{prefix}.{args.format}")
                else:
                    image.extract(prefix, fb_format=args.format)
            except pdfium.PdfiumError:
                traceback.print_exc()
            image.close()



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pypdfium2._library_scope
from pypdfium2.version import *
from pypdfium2._helpers import *
from pypdfium2 import raw, internal



# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import re
import sys
import json
import shutil
import hashlib
import urllib.request as url_request
from urllib.error import HTTPError

from base import *  # local


MIN_PDFIUM_FOR_VERIFY = 7415
MIN_GH_FOR_VERIFY = "2.47.0"

def _have_recent_gh():
    
    if not shutil.which("gh"):
        log("gh CLI is not installed")
        return False
    
    from packaging.version import Version
    gh_version = run_cmd(["gh", "--version"], cwd=None, capture=True)
    gh_version = Version( re.match(r"gh version ([\d.]+)", gh_version).group(1) )
    
    if gh_version >= Version(MIN_GH_FOR_VERIFY):
        return True
    else:
        log("gh CLI version is too old for verification")
        return False

def _get_gh_auth_state():
    return run_cmd(["gh", "auth", "status"], cwd=None, check=False).returncode == 0

def _gh_web_api(path):
    log(f"API Request {path}")
    with url_request.urlopen("https://api.github.com"+path) as h:
        return json.loads( h.read().decode() )

def _get_sha256sum(path):
    # https://stackoverflow.com/a/44873382/15547292
    if sys.version_info >= (3, 11):
        with open(path, "rb") as fh:
            return hashlib.file_digest(fh, "sha256").hexdigest()
    else:
        chunksize = 128 * 1024  # 128 KiB
        chunk = memoryview(bytearray(chunksize))
        hash = hashlib.sha256()
        
        with open(path, "rb", buffering=0) as fh:
            n = fh.readinto(chunk)
            while n:
                hash.update(chunk[:n])
                n = fh.readinto(chunk)
        
        return hash.hexdigest()


def _verify_impl(archives, pdfium_version, have_recent_gh):
    
    # https://github.com/cli/cli/issues/11803#issuecomment-3334820737
    
    if not have_recent_gh:
        raise SystemExit(f"--verify requires gh CLI >= {MIN_GH_FOR_VERIFY}")
    if pdfium_version < MIN_PDFIUM_FOR_VERIFY:
        raise SystemExit(f"--verify was passed, but the requested pdfium version is too low: {MIN_PDFIUM_FOR_VERIFY}")
    
    artifact_paths = tuple(archives.values())
    one_artifact = artifact_paths[0]
    
    is_authed = _get_gh_auth_state()
    if is_authed:
        log("gh is authenticated, using default verification")
        extra_args = []
    else:
        log("gh is not authenticated, attempting tokenless verification (may fail due to rate limit)")
        attest_path = DataDir/f"pdfium-binaries-{pdfium_version}-attestation.json"
        if not attest_path.exists():
            file_sum = _get_sha256sum(one_artifact)
            attest_json = _gh_web_api(f"/repos/bblanchon/pdfium-binaries/attestations/sha256:{file_sum}")
            attest_json = attest_json["attestations"][0]["bundle"]
            with attest_path.open("w") as fh:
                json.dump(attest_json, fh)
        extra_args = ["-b", str(attest_path)]
    
    verify_result = run_cmd(
        ["gh", "attestation", "verify", "-R", "bblanchon/pdfium-binaries", str(one_artifact), *extra_args, "--format=json"],
        cwd=DataDir, check=True, capture=True
    )
    log(f"{one_artifact.name}: verified by gh")
    
    # short circuit if there are no more archives - the typical setup caller downloads only one archive at a time
    if not len(archives) > 1:
        return
    
    verify_result = json.loads(verify_result)
    attested_artifacts = verify_result[0]["verificationResult"]["statement"]["subject"]
    attested_artifacts = {d["name"]: d["digest"]["sha256"] for d in attested_artifacts}
    
    for artifact in artifact_paths:
        exp_checksum = attested_artifacts[artifact.name]
        actual_checksum = _get_sha256sum(artifact)
        if exp_checksum == actual_checksum:
            log(f"{artifact.name}: verified by result checksum: {exp_checksum}")
        else:
            raise SystemExit(f"{artifact.name}: checksum mismatch: expected {exp_checksum} != actual {actual_checksum}")


def do_verify(verify, archives, version):
    
    have_recent_gh = _have_recent_gh()
    auto_verify = False
    if verify is None:
        verify = have_recent_gh and version >= MIN_PDFIUM_FOR_VERIFY
        auto_verify = verify
    
    if verify:
        try:
            _verify_impl(archives, version, have_recent_gh)
        except HTTPError as e:
            if auto_verify:
                log(f"Warning: Verification failed: {e}")
                log("Proceeding because it was auto-enabled")
            else:
                raise e
    else:
        log("Warning: Verification is off. If this is not intentional, make sure `gh` (GitHub CLI) is installed.")
